'''
Function:
    照片怀旧风格
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import numpy as np
from ..base import BaseBeautifier


'''照片怀旧风格'''
class NostalgicstyleBeautifier(BaseBeautifier):
    def __init__(self, **kwargs):
        super(NostalgicstyleBeautifier, self).__init__(**kwargs)
    '''迭代图片'''
    def iterimage(self, image):
        image = image.astype(np.float32)
        image_processed = image.copy()
        image_processed[..., 0] = image[..., 2] * 0.272 + image[..., 1] * 0.534 + image[..., 0] * 0.131
        image_processed[..., 1] = image[..., 2] * 0.349 + image[..., 1] * 0.686 + image[..., 0] * 0.168
        image_processed[..., 2] = image[..., 2] * 0.393 + image[..., 1] * 0.769 + image[..., 0] * 0.189
        image_processed[image_processed > 255.0] = 255.0
        return image_processed


'''initialize'''
from .glitch import GlitchBeautifier
from .cartoonise import CartooniseBeautifier
from .cartoongan import CartoonGanBeautifier
from .oilpainting import OilpaintingBeautifier
from .beziercurve import BezierCurveBeautifier
from .photomosaic import PhotomosaicBeautifier
from .characterize import CharacterizeBeautifier
from .douyineffect import DouyinEffectBeautifier
from .noteprocessor import NoteprocessorBeautifier
from .pencildrawing import PencilDrawingBeautifier
from .cartoonizeface import CartoonizeFaceBeautifier
from .nostalgicstyle import NostalgicstyleBeautifier
from .photocorrection import PhotocorrectionBeautifier
from .fastneuralstyletransfer import FastNeuralStyleTransferBeautifier
from .geneticfitting import GeneticFittingCircleBeautifier, GeneticFittingPolygonBeautifier


'''
Function:
    手写笔记处理
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import os
import cv2
import numpy as np
from PIL import Image
from ..base import BaseBeautifier
from scipy.cluster.vq import kmeans, vq


'''手写笔记处理'''
class NoteprocessorBeautifier(BaseBeautifier):
    def __init__(self, value_threshold=0.25, sat_threshold=0.20, num_colors=8, sample_fraction=0.05, white_bg=False, saturate=True, **kwargs):
        super(NoteprocessorBeautifier, self).__init__(**kwargs)
        self.num_colors = num_colors
        self.sample_fraction = sample_fraction
        self.value_threshold = value_threshold
        self.sat_threshold = sat_threshold
        self.white_bg = white_bg
        self.saturate = saturate
    '''迭代图片'''
    def iterimage(self, image):
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        sampled_pixels = self.getsampledpixels(image, self.sample_fraction)
        palette = self.getpalette(sampled_pixels)
        labels = self.applypalette(image, palette)
        if self.saturate:
            palette = palette.astype(np.float32)
            pmin = palette.min()
            pmax = palette.max()
            palette = 255 * (palette - pmin) / (pmax - pmin)
            palette = palette.astype(np.uint8)
        if self.white_bg:
            palette = palette.copy()
            palette[0] = (255, 255, 255)
        image_processed = Image.fromarray(labels, 'P')
        image_processed.putpalette(palette.flatten())
        image_processed.save('tmp.png', dpi=(300, 300))
        image_processed = cv2.imread('tmp.png')
        os.remove('tmp.png')
        return image_processed
    '''将调色板应用到给定的图像, 第一步是将所有背景像素设置为背景颜色, 之后是使用最邻近匹配将每个前景色映射到调色板中最接近的一个'''
    def applypalette(self, image, palette):
        bg_color = palette[0]
        fg_mask = self.getfgmask(bg_color, image)
        orig_shape = image.shape
        pixels = image.reshape((-1, 3))
        fg_mask = fg_mask.flatten()
        num_pixels = pixels.shape[0]
        labels = np.zeros(num_pixels, dtype=np.uint8)
        labels[fg_mask], _ = vq(pixels[fg_mask], palette)
        return labels.reshape(orig_shape[:-1])
    '''选取图像中固定百分比的像素, 以随机顺序返回'''
    def getsampledpixels(self, image, sample_fraction):
        pixels = image.reshape((-1, 3))
        num_pixels = pixels.shape[0]
        num_samples = int(num_pixels * sample_fraction)
        idx = np.arange(num_pixels)
        np.random.shuffle(idx)
        return pixels[idx[:num_samples]]
    '''提取采样的RGB值集合的调色板, 调色板第一个条目始终是背景色, 其余的是通过运行K均值聚类从前景像素确定的'''
    def getpalette(self, samples, return_mask=False, kmeans_iter=40):
        bg_color = self.getbgcolor(samples, 6)
        fg_mask = self.getfgmask(bg_color, samples)
        centers, _ = kmeans(samples[fg_mask].astype(np.float32), self.num_colors-1, iter=kmeans_iter)
        palette = np.vstack((bg_color, centers)).astype(np.uint8)
        if not return_mask: return palette
        return palette, fg_mask
    '''通过与背景颜色进行比较来确定一组样本中的每个像素是否为前景, 如果像素的值或饱和度与背景的阈值不同, 则像素被分类为前景像素'''
    def getfgmask(self, bg_color, samples):
        s_bg, v_bg = self.rgbtosv(bg_color)
        s_samples, v_samples = self.rgbtosv(samples)
        s_diff = np.abs(s_bg - s_samples)
        v_diff = np.abs(v_bg - v_samples)
        return ((v_diff >= self.value_threshold) | (s_diff >= self.sat_threshold))
    '''将RGB图像或RGB颜色数组转换为饱和度和数值, 每个都作为单独的32位浮点数组或值返回'''
    def rgbtosv(self, rgb):
        if not isinstance(rgb, np.ndarray): rgb = np.array(rgb)
        axis = len(rgb.shape) - 1
        cmax = rgb.max(axis=axis).astype(np.float32)
        cmin = rgb.min(axis=axis).astype(np.float32)
        delta = cmax - cmin
        saturation = delta.astype(np.float32) / cmax.astype(np.float32)
        saturation = np.where(cmax==0, 0, saturation)
        value = cmax / 255.0
        return saturation, value
    '''从图像或RGB颜色数组中获得背景颜色, 方法为通过将相似的颜色分组为相同的颜色并找到最常见的颜色'''
    def getbgcolor(self, image, bits_per_channel=6):
        assert image.shape[-1] == 3
        image_quantized = self.quantize(image, bits_per_channel).astype(int)
        image_packed = self.packrgb(image_quantized)
        unique, counts = np.unique(image_packed, return_counts=True)
        packed_mode = unique[counts.argmax()]
        return self.unpackrgb(packed_mode)
    '''减少给定图像中RGB三通道的位数'''
    def quantize(self, image, bits_per_channel=6):
        assert image.dtype == np.uint8
        shift = 8 - bits_per_channel
        halfbin = (1 << shift) >> 1
        return ((image.astype(int) >> shift) << shift) + halfbin
    '''将24位RGB三元组打包成一个整数, 参数rgb为元组或者数组'''
    def packrgb(self, rgb):
        orig_shape = None
        if isinstance(rgb, np.ndarray):
            assert rgb.shape[-1] == 3
            orig_shape = rgb.shape[:-1]
        else:
            assert len(rgb) == 3
            rgb = np.array(rgb)
        rgb = rgb.astype(int).reshape((-1, 3))
        packed = (rgb[:, 0] << 16 | rgb[:, 1] << 8 | rgb[:, 2])
        if orig_shape is None: return packed
        return packed.reshape(orig_shape)
    '''将一个整数或整数数组解压缩为一个或多个24位RGB值'''
    def unpackrgb(self, packed):
        orig_shape = None
        if isinstance(packed, np.ndarray):
            assert packed.dtype == int
            orig_shape = packed.shape
            packed = packed.reshape((-1, 1))
        rgb = ((packed >> 16) & 0xff, (packed >> 8) & 0xff, (packed) & 0xff)
        if orig_shape is None: return rgb
        return np.hstack(rgb).reshape(orig_shape + (3,))


'''
Function:
    照片油画化
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import cv2
import random
import numpy as np
from scipy import ndimage
from ..base import BaseBeautifier


'''照片油画化'''
class OilpaintingBeautifier(BaseBeautifier):
    def __init__(self, brush_width=5, palette=0, edge_operator='sobel', **kwargs):
        super(OilpaintingBeautifier, self).__init__(**kwargs)
        assert edge_operator in ['scharr', 'prewitt', 'sobel', 'roberts']
        self.brush_width = brush_width
        self.palette = palette
        self.edge_operator = edge_operator
    '''迭代图片'''
    def iterimage(self, image):
        # 计算图像梯度
        r = 2 * int(image.shape[0] / 50) + 1
        gx, gy = self.getgradient(cv2.cvtColor(image, cv2.COLOR_BGR2GRAY), (r, r), self.edge_operator)
        gh = np.sqrt(np.sqrt(np.square(gx) + np.square(gy)))
        ga = (np.arctan2(gy, gx) / np.pi) * 180 + 90
        # 画油画的所有位置
        canvas = cv2.medianBlur(image, 11)
        order = self.getdraworder(image.shape[0], image.shape[1], scale=self.brush_width * 2)
        # 画椭圆
        colors = np.array(image, dtype=np.float)
        for i, (y, x) in enumerate(order):
            length = int(round(self.brush_width + self.brush_width * gh[y, x]))
            if self.palette != 0: 
                color = np.array([round(colors[y, x][0] / self.palette) * self.palette + random.randint(-5, 5), \
                                  round(colors[y, x][1] / self.palette) * self.palette + random.randint(-5, 5), \
                                  round(colors[y, x][2] / self.palette) * self.palette + random.randint(-5, 5)], dtype=np.float)
            else:
                color = colors[y, x]
            cv2.ellipse(canvas, (x, y), (length, self.brush_width), ga[y, x], 0, 360, color, -1, cv2.LINE_AA)
        # 返回结果
        return canvas
    '''画油画的所有位置'''
    def getdraworder(self, h, w, scale):
        order = []
        for i in range(0, h, scale):
            for j in range(0, w, scale):
                y = random.randint(-scale // 2, scale // 2) + i
                x = random.randint(-scale // 2, scale // 2) + j
                order.append((y % h, x % w))
        return order
    '''prewitt算子'''
    def prewitt(self, img):
        img_gaussian = cv2.GaussianBlur(img, (3, 3), 0)
        kernelx = np.array([[1, 1, 1], [0, 0, 0], [-1, -1, -1]])
        kernely = np.array([[-1, 0, 1], [-1, 0, 1], [-1, 0, 1]])
        img_prewittx = cv2.filter2D(img_gaussian, -1, kernelx)
        img_prewitty = cv2.filter2D(img_gaussian, -1, kernely)
        return img_prewittx // 15.36, img_prewitty // 15.36
    '''roberts算子'''
    def roberts(self, img):
        roberts_cross_v = np.array([[0, 0, 0], [0, 1, 0], [0, 0, -1]])
        roberts_cross_h = np.array([[0, 0, 0], [0, 0, 1], [0, -1, 0]])
        vertical = ndimage.convolve(img, roberts_cross_v)
        horizontal = ndimage.convolve(img, roberts_cross_h)
        return vertical // 50.0, horizontal // 50.0
    '''利用边缘检测算子获得梯度'''
    def getgradient(self, img_o, ksize, edge_operator):
        if edge_operator == 'scharr':
            X = cv2.Scharr(img_o, cv2.CV_32F, 1, 0) / 50.0
            Y = cv2.Scharr(img_o, cv2.CV_32F, 0, 1) / 50.0
        elif edge_operator == 'prewitt':
            X, Y = self.prewitt(img_o)
        elif edge_operator == 'sobel':
            X = cv2.Sobel(img_o, cv2.CV_32F, 1, 0, ksize=5)  / 50.0
            Y = cv2.Sobel(img_o, cv2.CV_32F, 0, 1, ksize=5)  / 50.0
        elif edge_operator == 'roberts':
            X, Y = self.roberts(img_o)
        X = cv2.GaussianBlur(X, ksize, 0)
        Y = cv2.GaussianBlur(Y, ksize, 0)
        return X, Y


'''
Function:
    简单的照片矫正
Author:
    Charles
微信公众号:
    Charles的皮卡丘
'''
import cv2
import numpy as np
from ..base import BaseBeautifier
from imutils.perspective import four_point_transform


'''简单的照片矫正'''
class PhotocorrectionBeautifier(BaseBeautifier):
    def __init__(self, epsilon_factor=0.08, canny_boundaries=[100, 200], use_preprocess=False, **kwargs):
        super(PhotocorrectionBeautifier, self).__init__(**kwargs)
        self.epsilon_factor = epsilon_factor
        self.canny_boundaries = canny_boundaries
        self.use_preprocess = use_preprocess
    '''迭代图片'''
    def iterimage(self, image):
        # 预处理
        if self.use_preprocess:
            image_edge = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
            image_edge = cv2.GaussianBlur(image_edge, (5, 5), 0)
            image_edge = cv2.dilate(image_edge, cv2.getStructuringElement(cv2.MORPH_RECT, (3, 3)))
        else:
            image_edge = image.copy()
        image_edge = cv2.Canny(image_edge, self.canny_boundaries[0], self.canny_boundaries[1], 3)
        # 找到最大轮廓
        cnts = cv2.findContours(image_edge.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        cnts = cnts[0]
        if len(cnts) < 1: return image
        cnts = sorted(cnts, key=cv2.contourArea, reverse=True)
        for cnt in cnts:
            peri = cv2.arcLength(cnt, True)
            approx = cv2.approxPolyDP(cnt, self.epsilon_factor * peri, True)
            if len(approx) == 4: break
        if len(approx) != 4: return image
        # 矫正
        image_processed = four_point_transform(image, approx.reshape(4, 2))
        # 返回
        return image_processed